from __future__ import absolute_import
import solver4.libsolver4 as ls4
from ctypes import *
from ctypes import util


class SolverError(StandardError):
	def __init__(self, returnval, simulation):
		if returnval==ls4.SV_error_single_divergent:
			string = "Approximation of root of a single constraint is getting worse, not better."
		elif returnval==ls4.SV_error_group_divergent:
			string = "Approximation of solution to system is getting worse, not better."
		elif returnval==ls4.SV_error_single_maxiter:
			string = "Iterated too many times trying to find root of single constraint."
		elif returnval==ls4.SV_error_group_maxiter:
			string = "Iterated too many times trying to find solution to system."
		else:
			string = "Unknown exception from solver: %d" % returnval
			
		if returnval in [ls4.SV_error_single_divergent, ls4.SV_error_single_maxiter]:
			failed_constraint = ls4.SV_simulation_get_failed_constraint(simulation._sim)
			assert failed_constraint != None
			for c in simulation._constraints:
				if addressof(c._constraint.contents) == addressof(failed_constraint.contents):
					constraint = c
					break
			else:
				raise Exception, "Solver returned a failed-constraint error code but the " \
									"constraint is nowhere to be found."
		else:
			constraint = None
		
		self.string = string
		self.constraint = constraint
		StandardError.__init__(self, string, constraint)
	
	def __str__(self):
		return self.string


class Expr(object):
	"""An Expr object represents an expression in terms of Variable objects. Do not create
expressions using the constructor Expr(); instead, use the classmethods Expr.variable, 
Expr.constant, Expr.sum, Expr.product and Expr.exponential.

After an Expr object has been passed as an argument in the creation of another Expr object or in the 
creation of a Constraint, it is invalidated and should not be used."""
	
	@classmethod
	def variable(cls, var):
		"""Expr.variable(var) -> Expr\n
Creates a new expression object that evaluates to the value of the given variable."""
		return Expr(ls4.SV_expr_new_variable(var._var))
	
	@classmethod
	def constant(cls, const):
		"""Expr.constant(const) -> Expr\n
Creates a new expression object that evaluates to the given constant."""
		return Expr(ls4.SV_expr_new_constant(const))
	
	@classmethod
	def sum(cls, *addends):
		"""Expr.sum([...]) -> Expr\n
Creates a new expression object that evaluates to the sum of all of the arguments, which must be
Expr objects."""
		if len(addends)==0:
			return Expr.constant(0.0)
		else:
			sum = addends[0]._steal()
			for addend in addends[1:]:
				sum = ls4.SV_expr_new_sum(sum, addend._steal())
			return Expr(sum)
	
	@classmethod
	def product(cls, *factors):
		"""Expr.product([...]) -> Expr\n
Creates a new expression object that evaluates to the product of all of the arguments, which must be
Expr objects."""
		if len(factors)==0:
			return Expr.constant(1.0)
		else:
			product = factors[0]._steal()
			for factor in factors[1:]:
				product = ls4.SV_expr_new_product(product, factor._steal())
			return Expr(product)
	
	@classmethod
	def exponential(cls, power):
		"""Expr.exponential(power) -> Expr\n
Creates a new expression object that evaluates to 'e' raised to the given power."""
		return Expr(ls4.SV_expr_new_exponential(power._steal()))
	
	def __init__(self, _expr):
		assert isinstance(_expr, POINTER(ls4.SV_expr))
		self._expr = _expr
	
	def __del__(self):
		if self._expr:
			ls4.SV_expr_free(self._expr)
	
	@classmethod
	def _repr(cls, e, precedence=0):
		t = e.contents.type
		u = e.contents.u
		
		if t == ls4.SV_expr_type_sum:
			s = "%s + %s" % (cls._repr(u.sum.a, 0), cls._repr(u.sum.b, 0))
			if precedence>0:
				s = "("+s+")"
			return s
		
		elif t == ls4.SV_expr_type_product:
			s = "%s*%s" % (cls._repr(u.product.a, 1), cls._repr(u.product.b, 1))
			if precedence>1:
				s = "("+s+")"
			return s
		
		elif t == ls4.SV_expr_type_exponential:
			return "e^%s" % cls._repr(u.exponential.power, 2)
		
		elif t == ls4.SV_expr_type_variable:
			return "<var 0x%x>" % addressof(u.variable.var.contents)
		
		elif t == ls4.SV_expr_type_constant:
			return str(u.constant.constant)
		
		elif t == ls4.SV_expr_type_internal1 or t == ls4.SV_expr_type_internal2:
			return "<INTERNAL EXPR TYPE>"
		
		else:
			return "<UNKNOWN EXPR TYPE>"
	
	def represent(self):
		"""e.represent() -> str\n
Returns a string representation of the expression."""
		if not self._expr:
			raise RuntimeError, """This expression object is invalid, because it has become part of
another expression object or part of a Constraint object."""
		return self._repr(self._expr)
	
	def __repr__(self):
		if self._expr:
			return "<solver4.Expr: (%s)>" % self.represent()
		else:
			return "<solver4.Expr [invalid]>"
	
	def _steal(self):
		if not self._expr:
			raise RuntimeError, """This expression object is invalid, because it has become part of
another expression object or part of a Constraint object."""
		expr = self._expr
		self._expr = None
		return expr


class Variable(object):
	"""Variable() -> Variable\n
A Variable object represents a variable in the simulation."""
	def __init__(self):
		self._derivative = None
		self._var = ls4.SV_variable_new()
		
		self._in_simulation = False
	
	def __del__(self):
		ls4.SV_variable_free(self._var)
	
	def __repr__(self):
		return "<solver4.Variable %s>" % self.id
	
	def set_derivative(self, derivative):
		assert derivative==None or isinstance(derivative,Variable)
		self._derivative = derivative
		ls4.SV_variable_set_derivative(self._var, self._derivative._var)
	
	def get_derivative(self):
		return self._derivative
	
	derivative = property(get_derivative, set_derivative, None,
		"""v.derivative <-> Variable\n
The derivative of this variable, in the form of another variable. This property can be both read and
written.""")
	
	def get_id(self):
		return "0x%x" % addressof(self._var.contents)
	
	id = property(get_id, None, None,
		"""v.id -> string\n
Returns a string identifying the variable. This will be the same string used in the strings
returned by Expr.represent(). This property is read-only.""")


class Constraint(object):
	"""Constraint(expr) -> Constraint\n
A Constraint object represents a constraint on the values of a group of variables. Its main member
is an Expr object; the constraint is considered to be satisfied if the absolute value of this 
expression is less than some tolerance. When a constraint is created using the Constraint()
constructor, the expression given will be "owned" by the Constraint object, and the given Expr will
become invalid."""
	
	def __init__(self, expression):
		self._expr = expression._steal()
		self._constraint = ls4.SV_constraint_new(self._expr)
		
		self._in_simulation = False
	
	def __del__(self):
		ls4.SV_constraint_free(self._constraint)

	def __repr__(self):
		return "<solver4.Constraint: (%s)>" % Expr._repr(self._expr)

	def get_expr(self):
		return Expr(ls4.SV_expr_duplicate(self._expr))
	
	def set_expr(self, expression):
		self._expr = expression._steal()
		ls4.SV_constraint_replace_expr(self._constraint, self._expr)
	
	expr = property(get_expr, set_expr, None,
		"""c.expr <-> Expr\n
The expression that must be equal to 0. This property can be both read and written. If the property
is read, the result will be a duplicate of the actual expression. If this property is written, then
the expression given will be "owned" by the Constraint object, and the given Expr object will
become invalid.""")
	
	def get_property(self, prop, type):
		var = type()
		ls4.SV_constraint_get_property(self._constraint, prop, byref(var))
		return var.value
	
	def set_property(self, prop, type, value):
		var = type(value)
		ls4.SV_constraint_set_property(self._constraint, prop, byref(var))
	
	max_iterations = property(
		lambda self:   self.get_property(ls4.SV_constraint_property_max_iterations, c_int),
		lambda self,i: self.set_property(ls4.SV_constraint_property_max_iterations, c_int, i),
		None,
		"""c.max_iterations <-> int\n
If more than 'max_iterations' are necessary to solve the constraint, then the solver will give up
and an exception will be raised.""")
	
	tolerance = property(
		lambda self:   self.get_property(ls4.SV_constraint_property_tolerance, c_double),
		lambda self,t: self.set_property(ls4.SV_constraint_property_tolerance, c_double, t),
		None,
		"""c.tolerance <-> float\n
For a constraint to be considered 'satisfied', the absolute value of the constraint's expression
must be less than the constraint's tolerance.""")
	
	def get_id(self):
		return "0x%x" % addressof(self._constraint.contents)
	
	id = property(get_id, None, None,
		"""c.id -> string\n
Returns a string identifying the constraint. This string will be the hexadecimal address of the
underlying C constraint object in memory, which is useful for debugging.""")


class Simulation(object):
	"""Simulation() -> Simulation\n
A Simulation object represents a group of variables and constraints. Variables and constraints are
useless outside of a simulation."""
	def __init__(self):
		self._variables = []
		self._constraints = []
		
		self._sim = ls4.SV_simulation_new()
		
		self._debug_file = None
	
	def __del__(self):
		for var in self._variables:
			var._in_simulation = False
		for constraint in self._constraints:
			constraint._in_simulation = False
		ls4.SV_simulation_free(self._sim)
	
	def __repr__(self):
		return "<solver4.Simulation: %d variables, %d constraints>" % \
			(len(self._variables), len(self._constraints))
	
	def add_variable(self, var):
		"""s.add_variable(var) -> None\n
Makes the given variable part of the simulation. The variable must not be part of any other
simulation at the same time."""
		assert not var._in_simulation
		self._variables.append(var)
		var._in_simulation = True
		ls4.SV_simulation_add_variable(self._sim, var._var)
	
	def remove_variable(self, var):
		"""s.remove_variable(var) -> None\n
Removes the given variable from the simulation. The variable will still be "alive", and can be used
in another simulation."""
		self._variables.remove(var)
		var._in_simulation = False
		ls4.SV_simulation_remove_variable(self._sim, var._var)
	
	def get_variables(self):
		return iter(self._variables)
	
	variables = property(get_variables, None, None,
		"""s.variables -> iterator\n
Returns an iterator over all of the variables currently in the simulation.""")
	
	def add_constraint(self, constraint):
		"""s.add_constraint(constraint) -> None\n
Makes the given constraint part of the simulation. The constraint must not be part of any other
simulation at the same time, because the constraint must belong to the same simulation that the
variables it uses belong to, and the variables may only belong to one simulation at once."""
		assert not constraint._in_simulation
		self._constraints.append(constraint)
		constraint._in_simulation = True
		ls4.SV_simulation_add_constraint(self._sim, constraint._constraint)
	
	def remove_constraint(self, constraint):
		"""s.remove_constraint(constraint) -> None\n
Removes the given constraint from the simulation. The constraint will still be valid, and can be
used in another simulation."""
		self._constraints.remove(constraint)
		constraint._in_simulation = False
		ls4.SV_simulation_remove_constraint(self._sim, constraint._constraint)
	
	def get_constraints(self):
		return iter(self._constraints)
	
	constraints = property(get_constraints, None, None,
		"""s.constraints -> iterator\n
Returns an iterator over all of the constraints currently in the simulation.""")
	
	def get_property(self, prop, type):
		var = type()
		ls4.SV_simulation_get_property(self._sim, prop, byref(var))
		return var.value
	
	def set_property(self, prop, type, value):
		var = type(value)
		ls4.SV_simulation_set_property(self._sim, prop, byref(var))
	
	max_iterations = property(
		lambda self:   self.get_property(ls4.SV_simulation_property_max_iterations, c_int),
		lambda self,i: self.set_property(ls4.SV_simulation_property_max_iterations, c_int, i),
		None,
		"""s.max_iterations <-> int\n
If it takes the solver more than 'max_iterations' to find a solution to a system of equations, it
will give up and raise an error message.""")
	
	max_outer_timestep = property(
		lambda self:   self.get_property(ls4.SV_simulation_property_max_outer_timestep, c_double),
		lambda self,t: self.set_property(ls4.SV_simulation_property_max_outer_timestep, c_double,t),
		None,
		"""s.max_outer_timestep <-> float\n
'max_outer_timestep' is the largest amount by which the solver will increase time without 
recalculating the correlation matrix.""")

	max_inner_timestep = property(
		lambda self:   self.get_property(ls4.SV_simulation_property_max_inner_timestep, c_double),
		lambda self,t: self.set_property(ls4.SV_simulation_property_max_inner_timestep, c_double,t),
		None,
		"""s.max_inner_timestep <-> float\n
'max_inner_timestep' is the largest amount by which the solver will increase time in one chunk. 
Larger time increases will be broken up into chunks of 'max_inner_timestep' or smaller.
'max_inner_timestep' should be smaller than 'max_outer_timestep', although this is not mandatory. 
""")
	
	def calculate_completion(self, knowns, estimates={}):
		"""s.calculate_completion(knowns, estimate={}) -> dict\n
Given values for some or none of the variables in the simulation, finds values for the remainder.
'knowns' should be a dictionary that maps known variables to their values. 'estimate' may be a dict
that maps unknown variables to initial estimates of their value. If a variable is missing from
'estimate', then 0.0 will be used as an initial estimate. Return value is a dictionary that maps
the all variables that are not in 'knowns' to their values."""
		knowns_array = (c_int*len(self._variables))()
		variables_array = (c_double*len(self._variables))()
		
		for var in self._variables:
			var_id = ls4.SV_variable_get_id(var._var)
			if var in knowns:
				knowns_array[var_id] = 1
				variables_array[var_id] = knowns[var]
				assert var not in estimates
			elif var in estimates:
				variables_array[var_id] = estimates[var]
			else:
				variables_array[var_id] = 0.0
		
		result = ls4.SV_calculate_completion(self._sim, knowns_array, variables_array)
		
		if result==0:
			answers = {}
			for var in self._variables:
				if var not in knowns:
					var_id = ls4.SV_variable_get_id(var._var)
					answers[var] = variables_array[var_id]
			return answers
		
		else:
			raise SolverError(result, self)
	
	def calculate_change(self, variables, t):
		"""s.calculate_change(variables, time) -> dict\n
Given values for all of the variables at time 0, find the values of the variables after some amount
of time has passed. 'variables' should be a dictionary that maps Variable objects to numbers. 'time'
should be a number. The return value will be another dictionary that maps Variable objects to
numbers."""
		variables_array = (c_double*len(self._variables))()
		
		for var in self._variables:
			var_id = ls4.SV_variable_get_id(var._var)
			variables_array[var_id] = variables[var]
		
		result = ls4.SV_calculate_change(self._sim, variables_array, t)
		
		if result==0:
			answers = {}
			for var in self._variables:
				var_id = ls4.SV_variable_get_id(var._var)
				answers[var] = variables_array[var_id]
			return answers
		
		else:
			raise SolverError(result, self)
	
	def debug(self, include_names = False):
		"""s.debug([include_names]) -> None\n
Writes debugging information about the simulation to the simulation's debugging file in a format 
familiar to solver4debugger. This function must be called before performing any calculations that
are going to be recorded in the debug file.

If 'include_names' is True, then it will also include name-strings for the simulation and each
constraint and variable object. The name-string will be the 'name' property of the simulation,
constraint, or var if it has one; otherwise, it will be a string with the memory address of the
simulation, constraint, or var.
"""
		if not self._debug_file:
			return

		ls4.SV_debug_summarize_simulation(self._sim)
		
		if include_names:
			print >>self._debug_file, "===== simulation-names: %s =====" % self.id
			
			name = getattr(self, "name", self.id).encode("string-escape")
			print >>self._debug_file, "name = \"%s\"" % name
			
			for var in self._variables:
				print >>self._debug_file, "===== variable-name: %s =====" % var.id
				name = getattr(var, "name", self.id).encode("string-escape")
				print >>self._debug_file, "name = \"%s\"" % name
				print >>self._debug_file, "----- end -----"
			
			for constraint in self._constraints:
				print >>self._debug_file, "===== constraint-name: %s =====" % constraint.id
				name = getattr(constraint, "name", self.id).encode("string-escape")
				print >>self._debug_file, "name = \"%s\"" % name
				print >>self._debug_file, "----- end -----"
				
			print >>self._debug_file, "----- end -----"
	
	def set_debug_file(self, pfile):
		if pfile is None:
			self._debug_file = None
			p = c_int()
			p.value = 0
			ls4.SV_simulation_set_property(self._sim, SV_simulation_property_debug_file, byref(p))
		else:
			assert isinstance(pfile, file)
			self._debug_file = pfile
			cfile_address = pythonapi.PyFile_AsFile(py_object(pfile))
			cfile_buffer = pointer(c_int.from_address(cfile_address))
			ls4.SV_simulation_set_property(self._sim, ls4.SV_simulation_property_debug_file, 
											byref(cfile_buffer))
	
	def get_debug_file(self):
		return self._debug_file
	
	debug_file = property(get_debug_file, set_debug_file, None,
		"""s.debug_file <-> file object\n
If libsolver4 has been compiled with SV_DEBUG_LOGGING, then it can be configured to write debugging
information to a file. This file is specified by writing to the 'debug_file' property of a 
simulation object. The file given must be a Python file object opened for writing. If None is given,
then no debug information will be produced.

Once the debug file has been set, simulation.debug() must be called to write a summary of the 
simulation to the debug file. After this, subsequent calls to calculate_completion() or 
calculate_change() will automatically generate debugging information in the debug file.
If the simulation is modified, simulation.debug() must be called again before calling calculate_*()
again.""")
	
	def get_id(self):
		return "0x%x" % addressof(self._sim.contents)
	
	id = property(get_id, None, None,
		"""s.id -> string\n
Returns a string identifying the simulation. This string will be the hexadecimal address of the
underlying C simulation object in memory, which is useful for debugging.""")