from __future__ import division
from solver4 import *
import math

def close_enough(a,b):
	epsilon = 0.01
	return abs(a-b)<epsilon

# This test solves the following system of equations:
#
# a + b = 3*x [constraint 1]
# a = 2b      [constraint 2]
# x' = b
#
# This system reduces to x' = x.
#
# Then both constraints and the variables 'a' and 'b' are removed. The system is changed into:
#
# c + 2d = 4*x [constraint 3]
# c = 2d       [constraint 4]
# x' = d
#
# This system also reduces to x' = x.

s = Simulation()
s.max_outer_timestep = 0.01
s.max_inner_timestep = 0.001
s.max_iterations = 3 # Everything is linear, so this should be enough in theory.

# Create the first system

a = Variable()
b = Variable()
x = Variable()

C1 = Constraint(Expr.sum(Expr.variable(a), Expr.variable(b), Expr.product(Expr.variable(x), Expr.constant(-3))))
C2 = Constraint(Expr.sum(Expr.variable(a), Expr.product(Expr.variable(b), Expr.constant(-2))))

x.derivative = b

s.add_variable(a)
s.add_variable(b)
s.add_variable(x)
s.add_constraint(C1)
s.add_constraint(C2)

# Test the first system

vars = s.calculate_completion({x:1})
assert close_enough(vars[a], 2)
assert close_enough(vars[b], 1)

vars_t1 = s.calculate_change({x:1, a:2, b:1}, 1)
vars_t2 = s.calculate_change({x:1, a:2, b:1}, 2)

assert close_enough(vars_t1[x], math.e)
assert close_enough(vars_t2[x], math.e**2)

# Destroy the constraints and vars a and b.

s.remove_constraint(C1)
s.remove_constraint(C2)
s.remove_variable(a)
s.remove_variable(b)

# Create the second systme

c = Variable()
d = Variable()

C3 = Constraint(Expr.sum(Expr.variable(c),
						Expr.product(Expr.constant(2),Expr.variable(d)),
						Expr.product(Expr.constant(-4),Expr.variable(x))))
C4 = Constraint(Expr.sum(Expr.variable(c), Expr.product(Expr.constant(-2), Expr.variable(d))))

x.derivative = d

s.add_variable(c)
s.add_variable(d)
s.add_constraint(C3)
s.add_constraint(C4)

# Test the second system

vars = s.calculate_completion({x:1})
assert close_enough(vars[c], 2)
assert close_enough(vars[d], 1)

vars_t1 = s.calculate_change({x:1, c:2, d:1}, 1)
vars_t2 = s.calculate_change({x:1, c:2, d:1}, 2)
assert close_enough(vars_t1[x], math.e)
assert close_enough(vars_t2[x], math.e**2)